# coding: utf8

from __future__ import print_function
import numpy as np
import pandas as pd
import evaltools as evt
from sys import argv
from os.path import isdir, isfile
from distutils.dir_util import mkpath
from multiprocessing import Pool, Process
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib import cm
from datetime import datetime, date, timedelta, time

###################################
# read consol and namelist params #
###################################

model_list = ['model1', 'model2', 'model3']
sc = 'RMSE'

#################
# read listings #
#################

listing_path = "../doc/sample_data/listing"
stations = evt.utils.read_listing(
    listing_path, classes='all', types='all', area_coord=None)

##########
# params #
##########

modelColors = {'model1': '#00FFFF',
               'model2': '#3333FF',
               'model3': '#9E3BCF'}
modelMarkers = {'model1': '.',  # 'v',
                'model2': '.',  # '^',
                'model3': '.'}  # 'D'
modelMarkersTaylor = {'model1': 'v',
                      'model2': '^',
                      'model3': 'D'}
model_names = {'model1': 'ENSEMBLE MEDIAN',
               'model2': 'MOCAGE',
               'model3': 'CHIMERE'}
species_names = {'o3': 'Surface ozone',
                 'no2': 'Surface nitrogen dioxide',
                 'pm10': 'Surface PM10 aerosol',
                 'pm2p5': 'Surface PM2.5 aerosol',
                 'co': 'Surface carbon monoxide'}
score_names = {'RMSE': 'Root mean square error',
               'MMB': 'Modified mean bias',
               'PearsonR': 'Correlation',
               'Bias': 'Mean bias',
               'FracBias': 'Fractional bias'}
colors = [modelColors[model] for model in model_list]
mkpath("../doc/source/charts")

##########
# charts #
##########

print("quarterlyMedianScore")
for i in range(1, 4):
    last_quarter = evt.quarter.Quarter(
        startDate=date(2017, 1, 1),
        endDate=date(2017, 3, 31))
    ser = pd.Series(
        abs(np.random.normal(loc=i, size=6)),
        index=last_quarter.range(evt.quarter.Quarter.from_string("OND2015")),
        name='RMSE')
    ser.to_csv("../doc/source/charts/kpi{}.csv".format(i), sep=' ',
               na_rep='nan', float_format='%g', header=True, index=True,
               date_format='%Y%m%d%H')

evt.plotting.plot_quarterlyMedianScore(
    ["../doc/source/charts/kpi1.csv",
     "../doc/source/charts/kpi2.csv",
     "../doc/source/charts/kpi3.csv"],
    labels=['model1', 'model2', 'model3'],
    outputFile="../doc/source/charts/quarterlyMedianScore",
    title="Median of a score computed at every stations\n\n",
    colors=[modelColors[model] for model in model_list],
    linestyles=['-']*len(model_list),
    markers=[modelMarkers[model] for model in model_list],
    last_quarter=evt.quarter.Quarter(date(2017, 1, 1),
                                     date(2017, 3, 31)).string,
    first_quarter="OND2015",
    score='RMSE', thres=None, file_formats=['png'], ylabel=None,
    origin_zero=True, black_axes=True)


# objects building
obs_path = "../doc/sample_data/observations/{year}_{species}_{station}"
sim_path = ("../doc/sample_data/{model}{runtype}/J{forecastDay}/" +
            "{year}_{species}_{station}")
startDate = date(2017, 6, 1)
endDate = date(2017, 6, 4)
model_list = ["ENS", "MFM"]
model_colors = {"ENS": '#CB025A', "MFM": '#004D8A'}
species = 'no2'

observations = evt.evaluator.Observations.fromTimeSeries(
    generic_file_path=obs_path.format(year="{year}",
                                      species=species,
                                      station="{station}"),
    correc_unit=1e9,
    species=species,
    start=startDate,
    end=endDate,
    stations=stations,
    forecastHorizon=4
    )

objs = {}
dailyObjs = {}
fairmodeObjects = {}
fh = 4
for model in model_list:
    generic_file_path = sim_path.format(model=model, runtype='forecast',
                                        forecastDay='{forecastDay}',
                                        year='{year}', species=species,
                                        station='{station}')
    stationsIdx = observations.stations.index
    simulations = evt.evaluator.Simulations.fromTimeSeries(
        generic_file_path=generic_file_path,
        stationsIdx=stationsIdx,
        correc_unit=1,
        species=species,
        model=model,
        start=startDate,
        end=endDate,
        forecastHorizon=4,
        )
    objs[model] = evt.evaluator.Evaluator(observations, simulations,
                                          color=model_colors[model])
    dailyObjs[model] = objs[model].dailyMax()
    fairmodeObjects[model] = evt.fairmode.FairmodeEvaluator(
        observations, simulations, U=0.24, alpha=0.2, RV=200,
        beta=2, Np=5.2, Nnp=5.5, perc=.998, color=model_colors[model])

print("timeSeries")
evt.plotting.plot_timeSeries(
    list(objs.values()), station_list=stations.index[:3],
    colors=['#CB025A', '#004D8A'],
    outputFile="../doc/source/charts/timeSeries",
    start_end=[startDate, startDate+timedelta(days=3)],
    title="Time series of observed and modelled data\n\n",
    file_formats=['png'], black_axes=True, nb_of_minor_ticks=(7, 1))
evt.plotting.plot_timeSeries(
    list(objs.values()), station_list=stations.index[:3],
    colors=['#CB025A', '#004D8A'],
    outputFile="../doc/source/charts/timeSeries_quartiles",
    start_end=[startDate, startDate+timedelta(days=3)],
    title="Time series of observed and modelled data\n\n",
    file_formats=['png'], plot_type='median', envelope=True, black_axes=True)

print("meanTimeScores")
evt.plotting.plot_meanTimeScores(
    list(objs.values()), labels=objs.keys(),
    outputFile="../doc/source/charts/meanTimeScores_{score}",
    score_list=[sc],
    title="Mean time scores as a function of the forecast term\n",
    file_formats=['png'], black_axes=True)

print("medianStationScores")
evt.plotting.plot_medianStationScores(
    list(objs.values()), labels=objs.keys(), score_list=[sc],
    outputFile="../doc/source/charts/medianStationScores_{score}",
    title="Median station scores as a function of the forecast term\n",
    file_formats=['png'], black_axes=True)

print("timeScores")
evt.plotting.plot_timeScores(
    list(objs.values()), score_list=[sc], term=3,
    outputFile="../doc/source/charts/timeScores_{score}",
    title="Scores at a choosen forecast term\n", file_formats=['png'],
    xticking='auto', black_axes=True)

print("stationScores")
obj = list(objs.values())[0]
# evt.plotting.plot_stationScores(
#     obj, 'RMSE', forecastDay=0, title="", bbox=[-20, 50, 30, 70],
#     file_formats=['png'], point_size=20, higher_above=True,
#     outputFile="../doc/source/charts/stationScores")
evt.plotting.plot_stationScores(
    obj, 'RMSE', forecastDay=0, title="", bbox=[0., 25, 45, 55],
    file_formats=['png'], point_size=20, higher_above=True,
    outputFile="../doc/source/charts/stationScores_interp", interp2D=True)
evt.plotting.plot_stationScores(
    obj, 'RMSE', forecastDay=0, title="", bbox=[0., 25, 45, 55],
    file_formats=['png'], point_size=20, higher_above=True,
    land_color='#F0EEDC', sea_color='#8AB7DD', cmaplabel=r'RMSE $\mu g/m^3$',
    mark_by=('area', {'urb': 's', 'rur': 'o', 'sub': '^'}),
    outputFile="../doc/source/charts/stationScores")

print("taylorDiagram")
evt.plotting.plot_taylorDiagram(
    list(objs.values()), colors=['#CB025A', '#004D8A'], markers=['v', '^'],
    outputFile="../doc/source/charts/taylorDiagram",
    labels=objs.keys(), title="Taylor diagram", file_formats=['png'])


print("scoreQuartiles")
evt.plotting.plot_scoreQuartiles(
    list(objs.values()), xscore='FGE', yscore='SpearmanR',
    colors=['#CB025A', '#004D8A'], forecastDay=0,
    outputFile="../doc/source/charts/scoreQuartiles", labels=None,
    outputCSV=None, invert_xaxis=True, invert_yaxis=False, xmin=None,
    xmax=1, ymin=0, ymax=None, file_formats=['png'], black_axes=True)

print("comparisonScatterPlot")
evt.plotting.plot_comparisonScatterPlot(
    score='RMSE', forecastDay=0, xobject=list(objs.values())[0],
    yobject=list(objs.values())[1], title="Comparison of two simulations",
    outputFile="../doc/source/charts/comparisonScatterPlot", outputCSV=None,
    file_formats=['png'], black_axes=True)

print("significantDifferences")
persistence = observations.persistenceModel()
evt.plotting.plot_significantDifferences(
    score_list=['RMSE', 'FGE', 'SpearmanR', 'FactOf2'],
    former_objects=[list(objs.values())[0],
                    list(objs.values())[0],
                    list(objs.values())[1]],
    later_objects=[list(objs.values())[1], persistence, persistence],
    forecastDay=0, title="", xlabels=None,
    ylabels=["ENS/MFM", "ENS/pers.", "MFM/pers."],
    availability_ratio=0.75, file_formats=['png'],
    outputFile="../doc/source/charts/significantDifferences")

print("diurnalCycle")
evt.plotting.plot_diurnalCycle(
    list(objs.values()), station_list=stations.index[:3],
    outputFile="../doc/source/charts/diurnalCycle",
    title="Diurnal Cycle\n\n", file_formats=['png'], black_axes=True)
evt.plotting.plot_diurnalCycle(
    list(objs.values()), station_list=stations.index[:3],
    outputFile="../doc/source/charts/diurnalCycle_quartiles",
    title="Diurnal Cycle\n\n", file_formats=['png'], plot_type='median',
    envelope=True, black_axes=True)

print("stationScoreDensity")
evt.plotting.plot_stationScoreDensity(
    list(objs.values()), 'RMSE', forecastDay=0, title="",
    file_formats=['png'],
    outputFile="../doc/source/charts/stationScoreDensity")

print("dataDensity")
evt.plotting.plot_dataDensity(
    list(objs.values()), forecastDay=0, title="", file_formats=['png'],
    outputFile="../doc/source/charts/dataDensity")

# print("target")
# evt.fairmode.plot_targetDiagram(fairmodeObjects.values(),
#     file_formats=['png'], outputFile="../doc/source/charts/target")

# print("summary")
# fairmodeObjects['ENS'].plot_summary(availability_ratio=0.75, forecastDay=0,
#     outputFile="../doc/source/charts/summary", file_formats=['png'],
#     title="")

print("fairmode benchmark")
objs['ENS'].fairmodeBenchmark(
    targetFile="../doc/source/charts/target",
    summaryFile="../doc/source/charts/summary",)

print("tables.medianStationScores")
evt.tables.medianStationScores(
    list(objs.values()), forecastDay=0,
    score_list=['RMSE', 'MeanBias', 'PearsonR'], outputFile=None, labels=None,
    availability_ratio=0.75, min_nb_sta=10, float_format='%.2g')

print("tables.exceedancesScores")
evt.tables.exceedancesScores(
    dailyObjs.values(), forecastDay=0, thresholds=[20, 30, 40],
    outputFile=None, labels=None)

print("barScores")
evt.plotting.figsize = (8, 6)
evt.plotting.plot_barScores(
    objects=list(objs.values()),
    score_list=['RMSE', 'MeanBias'],
    forecastDay=0,
    averaging='median',
    title="Averaged station scores barplot",
    labels=None,
    colors=None,
    subregions=[['AT'], 'all'],
    xticksLabels=['Austria', 'all'],
    outputFile="../doc/source/charts/barScores",
    file_formats=['png'],
    availability_ratio=0.75,
    bar_kwargs={'rot': 0},
    )

print("barExceedances")
evt.plotting.plot_barExceedances(
    obj,
    threshold=20.,
    data="obs",
    start_end=None,
    forecastDay=0,
    labels=['Austria', 'others'],
    outputFile="../doc/source/charts/barExceedances",
    title="Daily threshold exceedances",
    ylabel="Number of exceedances",
    subregions=[['AT'], ['AD', 'CH', 'CZ']],
    ymin_max=None,
    bar_kwargs={'color': ['#E20B29', '#008C3C'],
                'width': .9},
    )

print("barContingencyTable")
evt.plotting.plot_barContingencyTable(
    list(objs.values()),
    threshold=20.,
    forecastDay=0,
    start_end=None,
    title="",
    outputFile="../doc/source/charts/barContingencyTable",
    bar_kwargs={},
    )

print("lineExceedances")
evt.plotting.plot_lineExceedances(
    list(dailyObjs.values()),
    threshold=20,
    start_end=None,
    forecastDay=0,
    labels=None,
    colors=None,
    linestyles=None,
    markers=None,
    outputFile="../doc/source/charts/lineExceedances",
    title="",
    ylabel="Number of incidences",
    file_formats=['png'],
    xticking='daily',
    date_format='%Y-%m-%d',
    ymin=None,
    ymax=None,
    obs_style={'color': 'k', 'alpha': 0.5},
    outputCSV=None)
