# coding: utf8
# Copyright (c) Météo France (2017-)
# This software is governed by the CeCILL-C license under French law.
# http://www.cecill.info
""" This module defines Dataset, a class which methods are
    designed to process input data from several formats. It also
    define Store, a class designed to store time series values
    in HDF5 format. """

from __future__ import print_function
import sys
import numpy as np
import pandas as pd
import evaltools as evt
from datetime import datetime, date, timedelta, time
from os.path import isdir, isfile
try:
    import h5py
except ImportError as e:
    pass


class Dataset(object):
    """ Dataset class for evaltools input data.

    This class is based on pandas.DataFrame class. The main attribute
    (data) is a pandas DataFrame with datetime index and a list of
    stations as columns.

    Parameters
    ----------
    stations : 1D array of str
        List of stations to keep in the returned Dataset.
    startDate : datetime.date object
        The date from which data is collected.
    endDate : datetime.date object
        The date until which data is collected.
    species : str
        Species to process.
    seriesType : str.
        It can be 'hourly' (values stored with a hourly timestep) or
        'daily' (values stored with a daily timestep).

    """
    def __init__(self, stations, startDate, endDate, species="",
                 seriesType='hourly'):
        # check arguments integrity
        if startDate > endDate:
            raise evt.EvaltoolsError(
                "startDate={startDate} must be <= endDate={endDate}".format(
                    startDate=startDate, endDate=endDate))

        # initialize attributes
        self.species = species
        self.nb_days = (endDate-startDate).days + 1
        self.startDate = startDate
        self.endDate = endDate

        # initialize dataframe according to the series type
        self.seriesType = seriesType
        if seriesType == 'hourly':
            startTime = datetime.combine(startDate, time(0))
            times = pd.DatetimeIndex([startTime + timedelta(hours=h)
                                      for h in range(self.nb_days*24)])
        elif seriesType == 'daily':
            times = pd.DatetimeIndex([self.startDate + timedelta(days=d)
                                      for d in range(self.nb_days)])
        else:
            raise evt.EvaltoolsError(
                "seriesType argument must be either 'hourly' or 'daily' !!!")
        self.data = pd.DataFrame(index=times, columns=stations, dtype=float)

    @property
    def date_format(self):
        return ('%Y%m%d%H' if self.seriesType is 'hourly' else '%Y%m%d')

    def nan_rate(self):
        """ Computed rate of missing values on in the whole dataframe. """
        nan_rate = (np.sum(np.sum(self.data.isna())) /
                    ((self.data.shape[0]*self.data.shape[1])*1.))
        return nan_rate

    def summary(self):
        """ Print summary statistics on the data.

        Nan rate, minimum and maximum are computed for each station. Then,
        the minimum, maximum and median of each of these statistics are
        displayed.

        """
        nanRate = (self.data.isna().sum()/self.data.shape[0]).quantile(
            [0, .5, 1])
        min_val = self.data.min().quantile([0, .5, 1])
        max_val = self.data.max().quantile([0, .5, 1])

        summary = pd.DataFrame({'nan rate': nanRate,
                                'min value': min_val,
                                'max value': max_val})
        summary.index.name = 'quantiles'
        print(summary)

    def add_stations_metadata(self, listing_path, **kwargs):
        """ Read station metadata.

        The first column of the listing file must contain the station codes.
        Metadata is saved in the attribute self.stations.

        Parameters
        ----------
        listing_path : str
            Path to the listing file containing metadata.
        **kwargs :
            These parameters (like 'sep', 'sub_list', ...) will be
            passed to evaltools.utils.read_listing().

        """
        stations = evt.utils.read_listing(listing_path, **kwargs)
        if not np.in1d(self.data.columns, stations.index).all():
            raise evt.EvaltoolsError(
                "Listing does not contain all stations of the Dataset !!!")
        self.stations = stations.loc[self.data.columns]

    def subPeriod(self, startDate, endDate):
        """ Build a new Dataset object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Dataset object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError("startDate must be >= {} !!!".format(
                self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError("endDate must be <= {} !!!".format(
                self.endDate))

        ds = self.__class__(stations=self.data.columns, startDate=startDate,
                            endDate=endDate, species=self.species,
                            seriesType=self.seriesType)
        ds.data.update(self.data)

        return ds

    def drop_unrepresentative_stations(self, availability_ratio=0.75,
                                       drop=True):
        """ List stations with a certain rate of missing values.

        Parameters
        ----------
        availability_ratio : float or None
            Minimal rate of data available on the period required to
            keep a station. If None, stations with only nan values
            are dropped.
        drop : bool
            If True, the dataset is modify inplace by dropping
            unrepresentative stations.

        Returns
        -------
            List of stations that do not fulfill the condition.

        """
        if availability_ratio is None:
            idx = (np.sum(~self.data.isna()) == 0.)
        else:
            idx = (np.sum(~self.data.isna())/self.data.shape[0] <
                   availability_ratio)
        dropped_sta = idx.index[idx]
        if drop is True:
            self.data.drop(columns=dropped_sta, inplace=True)
        return dropped_sta

    def read_values_from_csv(self, generic_file_path, correc_unit=1,
                             usecols=[0, 1, 6], sep=';'):
        """ Collect data from csv files.

        Modify Dataset.data in place using non-NA values from
        csv files with no header containing at least 3 columns:
        station codes, times and concentration values.

        Parameters
        ----------
        generic_file_path : str
            Path of the files with {year}, {month}, {day} instead
            of the year, month and day number, and {species} instead
            instead of the species name.
        correc_unit : float
            Multiplicative factor applied to original values.
        usecols : list of 3 integer
            List of column indices corresponding to ['station code', 'time',
            'value']. Column numbering begins at 0.
        sep : str
            Column separator used in the csv.

        """
        if self.seriesType != 'hourly':
            raise evt.EvaltoolsError("Dataset series type must be " +
                                     "'hourly' to use this method !!!")

        for d in range(-1, self.nb_days):  # -1 since an EEA csv contains
            # 1, 2, ..., 23 , 0 (of D+1)
            currentDate = self.startDate+timedelta(days=d)
            f_path = generic_file_path.format(
                species=self.species,
                year=currentDate.year,
                month=currentDate.strftime('%m'),
                day=currentDate.strftime('%d'))
            if isfile(f_path):
                df = pd.read_table(f_path,
                                   sep=';',
                                   names=['station', 'time', 'obs'],
                                   header=None,
                                   usecols=usecols)
                df.time = pd.to_datetime(df.time, format='%Y%m%d%H%M%S')
                df.drop_duplicates(inplace=True)
                try:
                    df = df.pivot(index='time',
                                  columns='station',
                                  values='obs')
                    if correc_unit != 1:
                        df = df*correc_unit
                    self.data.update(df)
                except ValueError as err:
                    print("*** Problem with file {file} (this file is "
                          "skipped): {err} ".format(file=f_path, err=err))
            else:
                print("*** Missing file: {}".format(f_path))

    read_EEA_obs_csv = read_values_from_csv

    def updateFromTimeSeries(self, generic_file_path, correc_unit=1):
        """ Update nan values of the current object with timeseries files.

        Timeseries files are text files where each lines contains a time
        (yyyymmddhh) or a date (yyyymmdd) and a value separated by spaces.

        Parameters
        ----------
        generic_file_path : str
            Generic path of timeseriesfiles with {year} instead of the year
            number and {station} instead of the station name.
        correc_unit : float
            Multiplicative factor applied to original values.

        """
        df = evt.timeseries.get_DF(
            stationsIdx=self.data.columns,
            generic_file_path=generic_file_path,
            start_date=self.startDate,
            end_date=self.endDate,
            lag=0,
            correc_unit=correc_unit,
            seriesType=self.seriesType)
        self.data.update(df)

    def updateFromDataset(self, dataset, correc_unit=1):
        """ Update Dataset with values from an other one.

        Update nan values with values from an other Dataset or
        from a pandas.DataFrame.

        Parameters
        ----------
        dataset : evaltools.dataset.Dataset or pandas.DataFrame
            Dataset where to take data.
        correc_unit : float
            Multiplicative factor applied to original values.

        """
        try:
            self.data.update(dataset.data*correc_unit)
        except AttributeError:
            self.data.update(dataset*correc_unit)

    def addNewStations(self, stations):
        """ Add new stations to the Dataset.

        Add new stations to the Dataset if they are not already
        present. Modify Dataset object in place.

        Parameters
        ----------
        stations : 1D array-like
            List of stations to add.

        """
        idx = ~np.in1d(stations, self.data.columns)
        new_stations = stations[idx]
        self.data = pd.concat(
            [self.data, pd.DataFrame(index=self.data.index,
                                     columns=new_stations)],
            axis=1,
            sort=False)

    def updateFromDaimonobsHdf5(self, file_path, instr_name, correc_unit=1,
                                add_new_stations=True):
        """ Update nan values of the current object with Daimonobs hdf5.

        The path where to find the data in the hdf5 file must be
        <instr_name>/<self.species>/Y

        Parameters
        ----------
        file_path : str
            Path of the hdf5 file.
        instr_name : str
            Name of the first order group of the hdf5 file where
            to find the data.
        correc_unit : float
            Multiplicative factor applied to original values.
        add_new_stations : bool
            If True, new stations found in the input file are added
            to the Dataset.

        """
        import daimonobs

        instr = daimonobs.DefInstrument(
            instr_name,
            "SurfConcInstrument",
            species=[self.species])
        instr.Read(file_path)

        df = pd.DataFrame({'lat': instr.lats,
                           'lon': instr.lons,
                           'time': instr.time,
                           'values': instr.conc[self.species]*correc_unit})

        metadata = None
        if metadata is None:
            df['code'] = ["{0:.2f}".format(lat) + "/" + "{0:.2f}".format(lon)
                          for lat, lon in zip(df['lat'], df['lon'])]

        df = df.pivot(index='time', columns='code', values='values')
        if self.seriesType == 'hourly':
            df.index = df.index.floor('H')
        elif self.seriesType == 'daily':
            df.index = df.index.floor('D')

        if not df.index.is_unique:
            df = df.groupby(df.index).mean()

        if add_new_stations is True:
            self.addNewStations(df.columns)

        self.data.update(df)

    def writeDaimonobsHdf5(self, file_path, instr_name):
        """ Write dataset in a hdf5 compatible with daimonobs.

        Parameters
        ----------
        file_path : str
            Path of the output file.
        instr_name : str
            Name of the first order group of the hdf5 file where
            to store the data.

        """
        import daimonobs

        instr = daimonobs.DefInstrument(
            instr_name,
            "SurfConcInstrument",
            species=[self.species])

        df = self.data
        df = df.stack().to_frame()
        df.reset_index(level=0, inplace=True)
        df.columns = ['time', 'conc']
        df = df.merge(self.stations, how='left', left_index=True,
                      right_index=True)

        nobs = df.shape[0]
        instr.nobs += nobs  # increase the instrument counter
        if nobs > 0:
            instr.conc[self.species] = np.ma.concatenate(
                (instr.conc[self.species],
                 np.ma.array(df.conc, dtype=instr.conc[self.species].dtype)))
            instr.lats = np.ma.concatenate(
                (instr.lats, np.ma.array(df.lat, dtype=instr.lats.dtype)))
            instr.lons = np.ma.concatenate(
                (instr.lons, np.ma.array(df.lon, dtype=instr.lons.dtype)))
            instr.time = np.ma.concatenate(
                (instr.time, np.ma.array(df.time, dtype=instr.np_typetime)))
            # if instr.hascov:
            #     instr.cov[self.species] = np.ma.concatenate(
            #         (instr.cov[self.species],
            #          np.ma.array(res['cov'],
            #                      dtype=instr.cov[self.species].dtype)))
            # By default all obs are accepted
            instr.mask = np.concatenate(
                (instr.mask, np.zeros((nobs), dtype=bool)))

        instr.Write(file_path)

    def to_txt(self, output_path):
        """ Write one txt file per station.

        Parameters
        ----------
        output_path : str
            Path of the output file. The Path must contain {station}
            instead of the station code.

        """
        for sta in self.data.columns:
            self.data[sta].to_csv(
                output_path.format(station=sta),
                sep=' ',
                na_rep='nan',
                float_format='%g',
                header=False,
                index=True,
                date_format=self.date_format)

    def check_threshold(self, threshold, drop=False, filePath=None):
        """ Check if values exceed a threshold.

        If there are values above the threshold, a message is printed
        and these values are set to nan if drop == True.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        drop : bool
            If True, values above the threshold are set to nan.
        filePath : None or str
            File path where to save the names of stations that exceed
            the threshold.

        """
        idx = (self.data > threshold)
        nb_val_above = np.sum(idx.values)
        if nb_val_above > 0:
            print(("*** {nb} values above threshold in the dataset !!! " +
                   "values are {action}.").format(
                       nb=nb_val_above,
                       action='kept'*(not drop)+'set to np.nan'*drop))
            if filePath is not None:
                self.data.stack().loc[idx.stack()].to_csv(
                    filePath, sep=' ', na_rep='nan', float_format='%g',
                    header=False, index=True, date_format=self.date_format)
            if drop is True:
                self.data[idx] = np.nan
        else:
            print("0 value above {} in the dataset.".format(threshold))


def deltaToolFormatting(Dataset_dict, outputDir):
    """ Write csv compatible with the JRC-DetaTool.

    Parameters
    ----------
    Dataset_dict : dictionary of evt.dataset.Dataset
        Keys of the dictionary must species known by DeltaTool.
    outputDir : str
        Directory where to write the files (one file by measurement site).

    """
    station_dataframes = {}
    for lab, ds in Dataset_dict.items():
        for sta in ds.data.columns:
            if sta in station_dataframes.keys():
                station_dataframes[sta][lab] = pd.DataFrame(
                    ds.data[sta].astype(float))
            else:
                station_dataframes[sta] = pd.DataFrame(
                    index=ds.data.index,
                    columns=['year', 'month', 'day', 'hour', lab])
                station_dataframes[sta].year = ds.data.index.year
                station_dataframes[sta].month = ds.data.index.month
                station_dataframes[sta].day = ds.data.index.day
                station_dataframes[sta].hour = ds.data.index.hour
                station_dataframes[sta][lab] = pd.DataFrame(
                    ds.data[sta].astype(float))
    for sta, df in station_dataframes.items():
        station_dataframes[sta].to_csv("{outputDir}/{station}.csv".format(
            outputDir=outputDir, station=sta), sep=';', na_rep='-999',
            float_format='%.5g', header=True, index=False)


class Store(object):
    """ Tool designed for storing hourly time series data in HDF5 files.

    A file created with this class has a special architecture. It can
    hold data from different species on a same time period. The time
    period covered has a fixed length but can be shifted day by day.
    Boundary dates are saved in the file attributes.

    For each species, several dataset can be stored containing data
    for a same station list that can be increased.

    Parameters
    ----------
    file_path : str
        Path of the hdf5 file to read. The file must exist.

    """
    char_dtype = h5py.special_dtype(vlen=str)

    def __init__(self, file_path, read_only=False):
        # open file
        if read_only is True:
            self.file = h5py.File(file_path, 'r')
        else:
            self.file = h5py.File(file_path, 'r+')

        # load dates
        self.startDate = date(int(self.file.attrs['start'][0:4]),
                              int(self.file.attrs['start'][4:6]),
                              int(self.file.attrs['start'][6:8]))
        self.endDate = date(int(self.file.attrs['end'][0:4]),
                            int(self.file.attrs['end'][4:6]),
                            int(self.file.attrs['end'][6:8]))
        self.nb_days = int((self.endDate - self.startDate).days + 1)

        try:
            self.seriesType = self.file.attrs['seriesType']
        except KeyError:
            self.file.attrs['seriesType'] = 'hourly'
            self.seriesType = self.file.attrs['seriesType']

        # nTimes is a number of hours (seriesType == 'hourly') or days
        # (seriesType == 'daily')
        if self.seriesType == 'hourly':
            self.nTimes = self.nb_days*24
        elif self.seriesType == 'daily':
            self.nTimes = self.nb_days
        else:
            raise evt.EvaltoolsError(
                "seriesType argument must be either 'hourly' or 'daily' !!!")

    @classmethod
    def newFile(cls, file_path, startDate, endDate, seriesType='hourly'):
        """ Build a Store object by writing a new hdf5 file.

        If the file already exists, it is truncated.

        Parameters
        ----------
        file_path : str
            Path of the file to create.
        startDate : datetime.date
            Starting date of the file data.
        endDate : datetime.date
            Ending date of the file data.

        Returns
        -------
            New Store object.

        """
        new_file = h5py.File(file_path, 'w')
        new_file.attrs['start'] = startDate.strftime('%Y%m%d')
        new_file.attrs['end'] = endDate.strftime('%Y%m%d')
        new_file.attrs['seriesType'] = seriesType
        return cls(file_path)

    def summary(self):
        """ Print summary statistics on the hdf5 file data.

        Nan rate, minimum and maximum are computed for each station. Then,
        the minimum, maximum and median of each of these statistics are
        displayed.

        """
        print("File = {}".format(self.file))
        print("Period = {startDate} - {endDate}".format(
            startDate=self.startDate, endDate=self.endDate))
        print("Time step = {}".format(self.seriesType))
        for species in self.file.keys():
            print("")
            print("---------- {} ----------".format(species))
            print("number of stations = {}".format(
                self.file[species]['stations']['code'].shape[0]))
            for model in self.file[species]['data'].keys():
                print("")
                print("* {} *".format(model))
                ds = self.getDataset(species=species,
                                     dsetName=model,
                                     startDate=self.startDate,
                                     endDate=self.endDate)
                ds.summary()

    def newSpecies(self, species, stations, dsetNames=[]):
        """ Add a new species to the file of the Store object.

        Parameters
        ----------
        species : str
            Species name that will be the name of the new hdf5 file group.
        stations : list of str
            List of the station codes corresponding to the columns of the
            datasets to be stored.
        dsetNames : None or list of str
            List of the dataset names to initialize. The new dataset will be
            filled with np.nan.

        """
        grp = self.file.create_group(species)
        grp.create_group('stations')
        grp['stations'].create_dataset('code',
                                       data=np.array(stations,
                                                     dtype=self.char_dtype),
                                       maxshape=(None,))
        grp.create_group('data')
        for name in dsetNames:
            self.newDataset(species, name)

    def newDataset(self, species, dsetName):
        """ Add a new dataset to the file of the Store object.

        Parameters
        ----------
        species : str
            Species name that refered to the file group name
            where to add the dataset.
        dsetName : str
            Name of the new dataset. The new dataset will be
            filled with np.nan.

        """
        if species not in self.file.keys():
            raise evt.EvaltoolsError(
                ("{} is not a species of the hdf5 file. Please use " +
                 " newSpecies method to add it.").format(species))

        grp = self.file[species]
        grp['data'].create_dataset(
            dsetName,
            dtype='float32',
            data=np.full((self.nTimes, grp['stations']['code'].shape[0]),
                         np.nan),
            compression="gzip",
            compression_opts=9,
            maxshape=(self.nTimes, None),
            fillvalue=np.nan)

    def addStations(self, species, new_stations):
        """ Add stations to the species dataframes.

        Add columns filled with np.nan to all the dataset
        of the species.
        New station codes must not already be present.

        Parameters
        ----------
        species : str
            Species name that refered to the file group name
            where to add the new stations.
        new_stations : list of str
            List of the codes of the new stations.

        """
        if np.in1d(new_stations,
                   self.file[species]['stations']['code'][:]).any():
            raise evt.EvaltoolsError(
                "Trying to add already existing stations !!!")

        stations = np.concatenate((self.file[species]['stations']['code'][:],
                                   new_stations))
        nsta = stations.shape[0]
        self.file[species]['stations']['code'].resize((nsta,))
        self.file[species]['stations']['code'][
            len(stations)-len(new_stations):] = np.array(new_stations).astype(
                '|S7')

        for dset in self.file[species]['data'].values():
            dset.resize((self.nTimes, nsta))

    def update(self, dataset, dsetName, add_new_stations=True):
        """ Update a dataset of the hdf5 file with Dataset object.

        Modify a dataset of the hdf5 file using non-NA values from passed
        Dataset object. The species of the modified dataset is the
        species of the Dataset object.

        Parameters
        ----------
        dataset : Dataset object
            Dataset object containing new values to add to the file.
        dsetName : str
            Name of the dataset to update.
        add_new_stations : bool
            If true, stations only present in dataset are also
            added to the hdf5 file.

        """
        # check if series types match
        if self.seriesType != dataset.seriesType:
            raise evt.EvaltoolsError(
                "Series types of dataset and self must be equal !!!")

        # check if h5 file holds the appropriate dataset
        species = dataset.species
        if species not in self.file.keys():
            raise evt.EvaltoolsError(
                ("{} is not a species of the hdf5 file. Please " +
                 "use newSpecies method to add it.").format(species))

        if dsetName not in self.file[species]['data'].keys():
            raise evt.EvaltoolsError(
                ("{dset} is not a dataset name of {species} " +
                 "group of the hdf5 file. Please use newDataset method to " +
                 "add it.").format(dset=dsetName, species=species))

        # find indices of data to load from the hdf5 file
        if self.startDate > dataset.startDate:
            first_idx = 0
            startTime = datetime.combine(self.startDate, time(0))
        else:
            first_idx = (dataset.startDate - self.startDate).days*24
            startTime = datetime.combine(dataset.startDate, time(0))

        if self.endDate < dataset.endDate:
            last_idx = self.nTimes
        else:
            last_idx = self.nTimes - (self.endDate - dataset.endDate).days*24

        # get station list from h5 file and add some if missing
        stations = self.file[species]['stations']['code'][:]
        if add_new_stations:
            new_stations = dataset.data.columns[
                np.logical_not(np.in1d(dataset.data.columns, stations))]
            self.addStations(species, new_stations)
            stations = self.file[species]['stations']['code'][:]

        # construct datetime index
        if self.seriesType == 'hourly':
            times = pd.DatetimeIndex([startTime+timedelta(hours=h)
                                      for h in range(last_idx-first_idx)])
        elif self.seriesType == 'daily':
            times = pd.DatetimeIndex([startTime.date()+timedelta(days=d)
                                      for d in range(last_idx-first_idx)])

        # get h5 file dataset
        df = pd.DataFrame(self.file[species]['data'][dsetName][
                              first_idx:last_idx, :],
                          index=times,
                          columns=stations)

        # write updated dataset in the h5 file
        df.update(dataset.data)
        self.file[species]['data'][dsetName][
            first_idx:last_idx, :] = df.values

    def shiftOneDay(self):
        """ Shift the hdf5 file time period one day.

        In all datasets of the file, rows corresponding to the
        earliest day are removed and new rows containing np.nan
        are addes for the new day.

        """
        for spe in self.file.keys():
            grp = self.file[spe]['data']
            for model in grp.keys():
                arr = grp[model][24:(24*self.nb_days)]
                grp[model][0:(24*(self.nb_days-1))] = arr
                grp[model][(24*(self.nb_days-1)):(24*self.nb_days)] = np.nan
        self.startDate = self.startDate + timedelta(days=1)
        self.endDate = self.endDate + timedelta(days=1)
        self.file.attrs['start'] = self.startDate.strftime('%Y%m%d')
        self.file.attrs['end'] = self.endDate.strftime('%Y%m%d')

    def getDataset(self, species, dsetName, startDate, endDate):
        """ Get a dataset of the hdf5 file.

        Parameters
        ----------
        species : str
            Species name (corresponding to the first wanted key of
            the hdf5 file).
        dsetName : str
            Name of the dataset to get.
        startDate : datetime.date object
            The date from which data is collected.
        endDate : datetime.date object
            The data until which data is collected.

        Returns
        -------
            evaltools.dataset.Dataset

        """
        # find indices of data to load from the hdf5 file
        if self.startDate > startDate:
            first_idx = 0
            startTime = datetime.combine(self.startDate, time(0))
        else:
            first_idx = (startDate - self.startDate).days*(
                24*(self.seriesType == 'hourly') +
                1*(self.seriesType == 'daily'))
            startTime = datetime.combine(startDate, time(0))
        if self.endDate < endDate:
            last_idx = self.nTimes
        else:
            last_idx = self.nTimes - (self.endDate - endDate).days*(
                24*(self.seriesType == 'hourly') +
                1*(self.seriesType == 'daily'))

        # construct datetime index
        if self.seriesType == 'hourly':
            times = pd.DatetimeIndex([startTime+timedelta(hours=h)
                                      for h in range(last_idx-first_idx)])
        elif self.seriesType == 'daily':
            times = pd.DatetimeIndex([startTime.date()+timedelta(days=d)
                                      for d in range(last_idx-first_idx)])

        # get DataFrame
        df = pd.DataFrame(self.file[species]['data'][dsetName][
                              first_idx:last_idx, :],
                          index=times,
                          columns=self.file[species]['stations']['code'])

        # construct Dataset
        res = Dataset(stations=df.columns,
                      startDate=startDate,
                      endDate=endDate,
                      species=species,
                      seriesType=self.seriesType)
        res.data.update(df)

        return res
