# coding: utf8
# Copyright (c) Météo France (2017-)
# This software is governed by the CeCILL-C license under French law.
# http://www.cecill.info

from __future__ import print_function
import numpy as np
import pandas as pd
import evaltools as evt
from datetime import datetime, date, timedelta, time
from os.path import isdir, isfile
import pickle
import copy
from functools import reduce
import re


class Observations(object):
    """ Class gathering observations of the studied case.

    An object of class Observations will be specific to one species,
    one series type and one period.

    Parameters
    ----------
    species : str
        Species name (ex: "o3").
    startDate : datetime.date
        Starting day of the studied period.
    endDate : datetime.date
        Ending day (included) of the studied period.
    stations : pandas.DataFrame
        DataFrame with station names as index.
    seriesType : str.
        It can be 'hourly' or 'daily'.
    forecastHorizon : int
        Number of day corresponding to the forecast horizon of the model.
        For example, if the forcast horizon is 4 days, the end date of the
        studied observations has to be 3 days further than the end of the
        studied period since the studied period corresponds to the period
        along which the model has been executed.

    """
    def __init__(self, species, startDate, endDate, stations, seriesType,
                 forecastHorizon=1):
        self.stations = stations
        self.forecastHorizon = forecastHorizon
        # main attribute
        self.dataset = evt.dataset.Dataset(
            stations=stations.index,
            startDate=startDate,
            endDate=endDate+timedelta(days=forecastHorizon-1),
            species=species,
            seriesType=seriesType)

    @property
    def startDate(self):
        return self.dataset.startDate

    @property
    def endDate(self):
        return self.dataset.endDate - timedelta(days=self.forecastHorizon-1)

    @property
    def species(self):
        return self.dataset.species

    @property
    def seriesType(self):
        return self.dataset.seriesType

    @classmethod
    def fromTimeSeries(cls, generic_file_path, species, start, end, stations,
                       correc_unit=1, seriesType='hourly', forecastHorizon=1,
                       availability_ratio=False):
        """ Class method used to construct an object from timeseries files.

        Parameters
        ----------
        generic_file_path : str
            Generic path of timeseriesfiles with {year} instead of
            the year number and {station} instead of the station name.
        species : str
            Species name (ex: "o3").
        start : datetime.date
            Starting day of the studied period.
        end : datetime.date
            Ending day (included) of the studied period.
        stations : pandas.DataFrame
            DataFrame with station names as index.
        correc_unit : float
            multiplicative factor applied to original values.
        seriesType : str
            Must be equal to 'hourly' if time series files contain one
            row per hour, or equal to 'daily' if they contain one row
            per day.
        forecastHorizon : int
            Number of day corresponding to the forecast horizon of the model.
            For example, if the forcast horizon is 4 days, the end date of
            the studied observations has to be 3 days further than the end
            of the studied period since the studied period corresponds to
            the period along which the model has been executed.
        availability_ratio : float or None
            Minimal rate of data available on the period required to
            keep a station. If None, stations with only nan values
            are dropped.

        Returns
        -------
            evaltools.evaluator.Observations object

        """
        obj = cls(seriesType=seriesType, species=species,
                  startDate=start, endDate=end, stations=stations,
                  forecastHorizon=forecastHorizon)
        obj.dataset.updateFromTimeSeries(generic_file_path, correc_unit)
        if availability_ratio is not False:
            obj.drop_unrepresentative_stations(availability_ratio)
        return obj

    @classmethod
    def fromDataset(cls, ds, listing_path, forecastHorizon=1, correc_unit=1,
                    **kwargs):
        """ Initialize from an evaltools.dataset.Dataset object.

        Parameters
        ----------
        ds : evaltools.dataset.Dataset object
            Dataset object corresponding observation data.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.
        **kwargs :
            These parameters (like 'sep', 'sub_list', ...) will be
            passed to evaltools.utils.read_listing().

        Returns
        -------
            evaltools.evaluator.Observations object

        """
        stations = evt.utils.read_listing(listing_path, **kwargs)
        if not np.in1d(ds.data.columns, stations.index).all():
            raise evt.EvaltoolsError(
                "Listing does not contain all stations of the Dataset !!!")
        stations = stations.loc[ds.data.columns]
        obj = cls(species=ds.species, startDate=ds.startDate,
                  endDate=ds.endDate-timedelta(days=forecastHorizon-1),
                  stations=stations, seriesType=ds.seriesType,
                  forecastHorizon=forecastHorizon)
        obj.dataset.updateFromDataset(ds, correc_unit)
        return obj

    @classmethod
    def fromHDF(cls, startDate, endDate, species, hdf_path, listing_path,
                forecastHorizon=1, correc_unit=1, **kwargs):
        """ Class method used to construct an object from HDF5 files.

        Parameters
        ----------
        startDate : datetime.date
            Starting day of the studied period.
        endDate : datetime.date
            Ending day (included) of the studied period.
        species : str
            Species name (must be a key of the hdf file).
        hdf_path : str
            Path of the input file. Input data can be splitted over
            two hdf files corresponding to two different years. In
            that case, the file path must contain '{year}'.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.
        **kwargs :
            These parameters (like 'sep', 'sub_list', ...) will be
            passed to evaltools.utils.read_listing().

        Returns
        -------
            evaltools.evaluator.Observations object

        """
        store_list = [evt.dataframe.Store(hdf_path.format(year=year),
                                          read_only=True)
                      for year in np.unique([startDate.year, endDate.year])]
        seriesType = store_list[0].seriesType
        ds_list = [store.getDF(
            species=species,
            dsetName='observations',
            startDate=startDate,
            endDate=endDate+timedelta(days=forecastHorizon-1))
                   for store in store_list]
        obs_ds = evt.dataset.Dataset(
            stations=reduce(np.union1d, [d.DF.columns for d in ds_list]),
            startDate=startDate,
            endDate=endDate+timedelta(days=forecastHorizon-1),
            species=species,
            seriesType=seriesType)
        for ds in ds_list:
            obs_ds.updateFromDataset(ds)
        observations = evt.evaluator.Observations.fromDataset(
            obs_ds, listing_path=listing_path,
            forecastHorizon=forecastHorizon, correc_unit=correc_unit,
            **kwargs)

        return observations

    def drop_unrepresentative_stations(self, availability_ratio=0.75):
        """ Drop stations with a certain rate of missing values.

        Modify Dataset object in place.

        Parameters
        ----------
        availability_ratio : float or None
            Minimal rate of data available on the period required to
            keep a station. If None, stations with only nan values
            are dropped.

        """
        self.dataset.drop_unrepresentative_stations(availability_ratio)
        self.stations = self.stations.loc[self.dataset.data.columns]

    def to_csv(self, filePath):
        """ Save timeseries dataframe as csv.

        Parameters
        ----------
        filePath : str
            Csv file path.

        """
        date_format = '%Y%m%d' + '%H'*(self.seriesType == 'hourly')
        self.obsDF.to_csv(filePath, sep=',', na_rep='nan', float_format='%g',
                          header=True, index=True, date_format=date_format)

    def check_values(self, threshold, drop=False, filePath=None):
        """ Check if observation values exceed a threshold.

        If there are values above the threshold, a message is printed
        and these values are set to nan if drop == True.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        drop : bool
            If True, values above the threshold are set to nan.
        filePath : None or str
            File path where to save the names of stations that exceed
            the threshold.

        """
        print("Observations threshold exceedances check:")
        self.dataset.check_threshold(threshold, drop, filePath)

    def persistenceModel(self, color='k'):
        """ Build an Evaluator object based on persistence model.

        Parameters
        ----------
        color : str
            Default color that will be used in plotting functions
            for the new object.

        Returns
        -------
            Evaluator object.

        """
        simulations = Simulations(
            startDate=self.startDate+timedelta(days=1),
            endDate=self.endDate,
            stations=self.stations.index,
            species=self.species,
            model="persistence",
            seriesType=self.seriesType,
            forecastHorizon=self.forecastHorizon)
        for fday in range(self.forecastHorizon):
            df = pd.DataFrame(
                self.dataset.data.values,
                columns=self.dataset.data.columns,
                index=self.dataset.data.index+timedelta(days=fday+1))
            simulations.datasets[fday].updateFromDataset(df)
        return Evaluator(
            self.subPeriod(self.startDate+timedelta(days=1), self.endDate),
            simulations,
            color)

    def subPeriod(self, startDate, endDate):
        """ Build a new Observations object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Observations object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError(
                "startDate must be >= {} !!!".format(self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError(
                "endDate must be <= {} !!!".format(self.endDate))

        new_obs = self.__class__(
            species=self.species, startDate=startDate, endDate=endDate,
            stations=self.stations, seriesType=self.seriesType,
            forecastHorizon=self.forecastHorizon)
        new_obs.dataset.updateFromDataset(self.dataset)

        return new_obs

    def simVSobs(self, grid, time, point_size=20, vmin=None, vmax=None,
                 cmap=None, colors=None, bounds=None, outputFile=None,
                 file_formats=['png']):
        """ Scatter plot of observations above simulation raster.

        Parameters
        ----------
        grid : evaltools.interpolation.Grid
            Grid object that must contain data for the species of
            the current Observations object.
        time : datetime.datetime or datetime.date
            Time for which to plot the observation values. This time
            must be contained in the current Observations object.
        point_size : float
            Point size (as define in matplotlib.pyplot.scatter).
        vmin, vmax : None or scalar
            Min and max values for the legend colorbar. If None, these
            values are found automatically.
        cmap : None or matplotlib.colors.Colormap object
            Colors used for plotting (default: matplotlib.cm.jet).
        colors : None or list of str
            List of color used in the chart if you want to discretize
            the values.
        bounds : None or list of scalar
            Boundary values for each category if you want to discretize
            the values. Arguments vmin and vmax must not be None, and the
            boundary values contained between vmin and vmax.
            Ignored if colors is None.
        outputFile : None str
            File where to save the plot (without extension).
            If None, the figure is shown in a popping window.
        file_formats : list of str
            List of file extensions.

        """
        import matplotlib.pyplot as plt
        from matplotlib import cm
        from mpl_toolkits.basemap import Basemap
        import matplotlib as mpl

        # define default colormap
        if cmap is None:
            cmap = cm.jet

        # plotting
        plt.close('all')
        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111)

        # draw coast lines with Basemap
        m = Basemap(resolution='i', projection='merc',
                    llcrnrlat=grid.minLat, urcrnrlat=grid.maxLat,
                    llcrnrlon=grid.minLon, urcrnrlon=grid.maxLon,
                    lat_ts=(grid.minLat + grid.maxLat)/2)
        m.drawcoastlines(linewidth=0.5, color='k', zorder=1)

        # discretizing values if colors is not None
        norm = None
        if colors is not None:
            cmap = mpl.colors.ListedColormap(colors)
            if bounds is not None:
                bounds = [vmin] + bounds + [vmax]
                norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        # plot simulation raster
        m.imshow(
            grid.grids[self.species],
            extent=(grid.minLon, grid.maxLon, grid.minLat, grid.maxLat),
            zorder=0,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            norm=norm)

        # plot observation values
        m.scatter(
            list(self.stations['lon']),
            list(self.stations['lat']),
            c=self.obsDF.loc[time],
            latlon=True,
            edgecolors='k',
            linewidths=1,
            s=point_size,
            zorder=2,
            cmap=cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax
            )
        cb = m.colorbar(ticks=bounds)

        # save figure
        evt.plotting._save_figure(outputFile, file_formats)

    def dailyMean(self, availability_ratio=0.75):
        """ Build Observations object working on daily mean.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily mean.

        Returns
        -------
        evaluator.Observations
            Observations object with seriesType = 'daily' and with data
            corresponding to the computed daily mean.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_obs = self.__class__(
            species=self.species, startDate=self.startDate,
            endDate=self.endDate, stations=self.stations,
            seriesType='daily', forecastHorizon=self.forecastHorizon)
        new_obs.dataset.updateFromDataset(
            evt.timeseries.daily_mean(self.dataset.data, availability_ratio))
        return new_obs

    def dailyMax(self, availability_ratio=0.75):
        """ Return Observations object working on daily maximum.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily maximum.

        Returns
        -------
        evaluator.Observations
            Observations object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_obs = self.__class__(
            species=self.species, startDate=self.startDate,
            endDate=self.endDate, stations=self.stations,
            seriesType='daily', forecastHorizon=self.forecastHorizon)
        new_obs.dataset.updateFromDataset(
            evt.timeseries.daily_max(self.dataset.data, availability_ratio))
        return new_obs

    def movingAverageDailyMax(self, availability_ratio=0.75):
        """ Compute the daily maximum of the 8-hour moving average.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available to compute the average for a
            given 8-hour window, and also minimal rate of values available
            for a given day to compute the daily maximum of the moving
            average. For example with availability_ratio=0.75, a daily maximum
            eight hours average can only be calculated if 18 eight hours
            average are available each of which requires 6 hourly values to
            be available.

        Returns
        -------
        evaluator.Observations
            Observations object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum of the moving average.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_obs = self.__class__(
            species=self.species, startDate=self.startDate,
            endDate=self.endDate, stations=self.stations,
            seriesType='daily', forecastHorizon=self.forecastHorizon)

        temp_df = copy.deepcopy(self.dataset.data)
        # add nan for the day before since 1st window is 17 to 1 o'clock
        for new_row in temp_df.index[0:24]-timedelta(days=1):
            temp_df.loc[new_row] = np.nan
        temp_df.sort_index(inplace=True)
        temp_df = evt.timeseries.moving_average_daily_max(
            temp_df, availability_ratio)

        new_obs.dataset.updateFromDataset(temp_df)

        return new_obs

    def filteredSeries(self, availability_ratio=0.75):
        """ Return Observations object working on the filtered series.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available for a given hour to compute the
            daily filtered series for this hour in all days.

        Returns
        -------
        evaluator.Observations
            Observations object with seriesType = 'hourly' and with data
            corresponding to the computed filtered series.

        """
        if self.seriesType is not 'hourly':
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_obs = self.__class__(
            species=self.species, startDate=self.startDate,
            endDate=self.endDate, stations=self.stations,
            seriesType='hourly', forecastHorizon=self.forecastHorizon)
        new_obs.dataset.updateFromDataset(
            evt.timeseries.filtered_series(self.dataset.data,
                                           availability_ratio))
        return new_obs

    def normalized_series(self):
        """ Return Observations object with normalized series.

        This method normalizes each series of observations by substracting
        its median and dividing by its interquartile range.

        Returns
        -------
        evaluator.Observations
            Observations object  with data corresponding to the computed
            normalized series.

        """
        new_obs = self.__class__(
            species=self.species, startDate=self.startDate,
            endDate=self.endDate, stations=self.stations,
            seriesType=self.seriesType, forecastHorizon=self.forecastHorizon)
        new_obs.dataset.updateFromDataset(
            evt.timeseries.normalized_series(self.dataset.data))
        return new_obs


class Simulations(object):
    """ Class gathering simulations of the studied case.

    An object of class Simulations will be specific to one species,
    one series type, one period and one model.

    Parameters
    ----------
    startDate : datetime.date
        Start day of the studied period.
    endDate : datetime.date
        End day (included) of the studied period.
    stations : 1D array-like of str
        List of the names of studied stations.
    species : str
        Species name (ex: "o3").
    model : str
        Name of the model that produced the simulated data.
    seriesType : str
        It can be 'hourly' or 'daily'.
    forecastHorizon : int
        Number of day corresponding to the forcast horizon of the model.

    """
    def __init__(self, startDate, endDate, stations, species, model,
                 seriesType, forecastHorizon):
        self.species = species
        self.startDate = startDate
        self.endDate = endDate
        self.forecastHorizon = forecastHorizon
        self.model = model
        self.seriesType = seriesType
        # main attribute
        self.datasets = []
        for fd in range(forecastHorizon):
            self.datasets.append(
                evt.dataset.Dataset(
                    stations=stations,
                    startDate=startDate+timedelta(days=fd),
                    endDate=endDate+timedelta(days=fd),
                    species=species,
                    seriesType=seriesType))

    @property
    def stations(self):
        return self.datasets[0].data.columns

    @classmethod
    def fromTimeSeries(cls, generic_file_path, stationsIdx, species, model,
                       start, end, forecastHorizon=1, correc_unit=1,
                       seriesType='hourly', availability_ratio=False):
        """ Class method used to construct an object from timeseries files.

        Parameters
        ----------
        generic_file_path : str
            Generic path of timeseriesfiles with {year} instead of
            the year number, {station} instead of the station name
            and {forecastDay} instead of the forecast day number.
        stationsIdx : list of str
            List of the names of studied stations.
        species : str
            Species name (ex: "o3").
        model : str
            Name of the model that produced the simulated data.
        start : datetime.date
            Start day of the studied period.
        end : datetime.date
            End day (included) of the studied period.
        forecastHorizon : int
            Number of day corresponding to the forcast horizon of the model.
        correc_unit : float
            Multiplicative factor applied to original values.
        seriesType : str
            It can be 'hourly' or 'daily'.
        availability_ratio : float or None
            Minimal rate of data available on the period required to
            keep a station. If None, stations with only nan values
            are dropped.

        """
        obj = cls(startDate=start, endDate=end, stations=stationsIdx,
                  species=species, model=model, seriesType=seriesType,
                  forecastHorizon=forecastHorizon)
        for fd in range(forecastHorizon):
            obj.datasets[fd].updateFromTimeSeries(
                generic_file_path.format(year="{year}",
                                         station="{station}",
                                         forecastDay=fd),
                correc_unit)
        if availability_ratio is not False:
            obj.drop_unrepresentative_stations(availability_ratio)
        return obj

    @classmethod
    def fromDataset(cls, model, ds_list, stationsIdx, correc_unit=1):
        """ Initialize from an evaltools.dataset.Dataset object.

        Stations kept in the returned Dataset are the intersection
        of stations from the Dataset objects of ds_list.

        Parameters
        ----------
        model : str
            Name of the model.
        ds_list : list of evaltools.dataset.Dataset objects
            Dataset objects corresponding to the different
            forecast terms. The order in the list matters.
        stationsIdx : list of str
            List of the names of studied stations.
        correct_unit : float
            Factor to apply to original values.

        Returns
        -------
            evaltools.evaluator.Simulations object

        """
        # define station list as the intersection of stations of the datasets
        stations = reduce(np.intersect1d,
                          [ds.data.columns for ds in ds_list] + [stationsIdx])

        # check metadata
        forecastHorizon = len(ds_list)
        species = ds_list[0].species
        startDate = ds_list[0].startDate
        endDate = ds_list[0].endDate
        seriesType = ds_list[0].seriesType
        for fd, ds in enumerate(ds_list):
            if (species != ds.species or
                    ds.startDate != (startDate+timedelta(days=fd)) or
                    ds.endDate != (endDate+timedelta(days=fd)) or
                    seriesType != ds.seriesType):
                raise evt.EvaltoolsError("Metadata of Dataset objects from " +
                                         "ds_list are not compatible !!!")

        # build Simulations object
        obj = cls(startDate=startDate, endDate=endDate,
                  stations=stations, species=species, model=model,
                  seriesType=seriesType, forecastHorizon=forecastHorizon)
        for fd, ds in enumerate(ds_list):
            obj.datasets[fd].updateFromDataset(ds, correc_unit)

        return obj

    @classmethod
    def fromHDF(cls, startDate, endDate, species, model, stations, hdf_path,
                stationsIdx, forecastHorizon=1, correc_unit=1):
        """ Class method used to construct an object from HDF5 files.

        Parameters
        ----------
        startDate : datetime.date
            Starting day of the studied period.
        endDate : datetime.date
            Ending day (included) of the studied period.
        species : str
            Species name (must be a key of the hdf file).
        model : str
            Name of the model.
        stations : 1D array of str
            List of stations to keep in the resulting object.
        hdf_path : str
            Path of the input file. The file path must contain
            '{forecastDay}' instead of the forecast day of the file.
            Input data can be splitted over two hdf files
            corresponding to two different years. In that case, the
            file path must contain '{year}'.
        stationsIdx : list of str
            List of the names of studied stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.
        **kwargs :
            These parameters (like 'sep', 'sub_list', ...) will be
            passed to evaltools.utils.read_listing().

        Returns
        -------
            evaltools.evaluator.Simulations object

        """
        sim_ds_list = []
        for fd in range(forecastHorizon):
            store_list = [evt.dataframe.Store(hdf_path.format(year=y,
                                                              forecastDay=fd),
                                              read_only=True)
                          for y in np.unique([startDate.year, endDate.year])]
            ds_list = [store.getDF(species=species,
                                   dsetName=model,
                                   startDate=startDate+timedelta(days=fd),
                                   endDate=endDate+timedelta(days=fd))
                       for store in store_list]
            ds = evt.dataset.Dataset(
                stations=np.intersect1d(reduce(np.union1d,
                                               [ds.data.columns
                                                for ds in ds_list]),
                                        stations),
                startDate=startDate+timedelta(days=fd),
                endDate=endDate+timedelta(days=fd),
                species=species)
            for temp_ds in ds_list:
                ds.updateFromDataset(temp_ds)
            sim_ds_list.append(ds)

        simulations = evt.evaluator.Simulations.fromDataset(
            model=model,
            ds_list=sim_ds_list,
            stationsIdx=stationsIdx,
            correc_unit=correc_unit)

        return simulations

    def drop_unrepresentative_stations(self, availability_ratio=0.75):
        """ Drop stations with a certain rate of missing values.

        Modify Dataset object in place. If for every forecast days,
        the condition is not fulfilled, a station is dropped.

        Parameters
        ----------
        availability_ratio : float or None
            Minimal rate of data available on the period required to
            keep a station. If None, stations with only nan values
            are dropped.

        """
        dropped_sta = np.array([])
        for ds in self.datasets:
            ustations = ds.drop_unrepresentative_stations(availability_ratio,
                                                          drop=False)
            dropped_sta = np.union1d(dropped_sta, ustations)

        for ds in self.datasets:
            ds.data.drop(columns=dropped_sta, inplace=True)

    def to_obs(self, forecastDay=0):
        """ Transform Simulations object to Observations object.

        Parameters
        ----------
        forecastDay : int
            Forecast day for which to keep simulation data.

        """
        stationsDF = pd.DataFrame(index=self.stations)
        obs = Observations(self.species, self.startDate,
                           self.endDate, stationsDF, self.seriesType,
                           self.forecastHorizon)
        obs.dataset.updateFromDataset(self.datasets[forecastDay])
        return obs

    def to_csv(self, filePath):
        """ Save timeseries dataframe as csv.

        Parameters
        ----------
        filePath : str
            File path with {forecastDay} instead of forecast day number.

        """
        date_format = '%Y%m%d' + '%H'*(self.seriesType == 'hourly')
        for fd in range(self.forecastHorizon):
            self.simDF[fd].to_csv(
                filePath.format(forecastDay=fd), sep=',', na_rep='nan',
                float_format='%g', header=True, index=True,
                date_format=date_format)

    def subPeriod(self, startDate, endDate):
        """ Build a new Simulations object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Simulations object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError(
                "startDate must be >= {} !!!".format(self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError(
                "endDate must be >= {} !!!".format(self.endDate))

        # select sub-period
        new_sim = self.__class__(
            startDate=startDate, endDate=endDate, stations=self.stations,
            species=self.species, model=self.model,
            seriesType=self.seriesType, forecastHorizon=self.forecastHorizon)
        for fd in range(self.forecastHorizon):
            new_sim.datasets[fd].updateFromDataset(self.datasets[fd])
        return new_sim

    def check_values(self, threshold, drop=False, filePath=None):
        """ Check if observation values exceed a threshold.

        If there are values above the threshold, a message is printed
        and these values are set to nan if drop == True.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        drop : bool
            If True, values above the threshold are set to nan.
        filePath : None or str
            File path where to save the names of stations that exceed
            the threshold, the path must contain {forecastDay} instead
            of the forecast day number.

        """
        for fd in range(self.forecastHorizon):
            print("{} threshold exceedances check ".format(self.model) +
                  "in forecast day = {}:".format(fd))
            self.datasets[fd].check_threshold(
                threshold, drop, filePath.format(forecastDay=fd))

    def dailyMean(self, availability_ratio=0.75):
        """ Build Simulations object working on daily mean.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily mean.

        Returns
        -------
        evaluator.Simulations
            Simulations object with seriesType = 'daily' and with data
            corresponding to the computed daily mean.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_sim = self.__class__(
            startDate=self.startDate, endDate=self.endDate,
            stations=self.stations, species=self.species,
            model=self.model, seriesType='daily',
            forecastHorizon=self.forecastHorizon)
        for fd in range(self.forecastHorizon):
            new_sim.datasets[fd].updateFromDataset(
                evt.timeseries.daily_mean(
                    self.datasets[fd].data, availability_ratio))
        return new_sim

    def dailyMax(self, availability_ratio=0.75):
        """ Return Simulations object working on daily maximum.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily maximum.

        Returns
        -------
        evaluator.Simulations
            Simulations object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_sim = self.__class__(
            startDate=self.startDate, endDate=self.endDate,
            stations=self.stations, species=self.species,
            model=self.model, seriesType='daily',
            forecastHorizon=self.forecastHorizon)
        for fd in range(self.forecastHorizon):
            new_sim.datasets[fd].updateFromDataset(
                evt.timeseries.daily_max(
                    self.datasets[fd].data, availability_ratio))
        return new_sim

    def movingAverageDailyMax(self, availability_ratio=0.75):
        """ Compute the daily maximum of the 8-hour moving average.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available to compute the average for a
            given 8-hour window, and also minimal rate of values available
            for a given day to compute the daily maximum of the moving
            average. For example with availability_ratio=0.75, a daily maximum
            eight hours average can only be calculated if 18 eight hours
            average are available each of which requires 6 hourly values to
            be available.

        Returns
        -------
        evaluator.Simulations
            Simulations object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum of the moving average.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_sim = self.__class__(
            startDate=self.startDate, endDate=self.endDate,
            stations=self.stations, species=self.species,
            model=self.model, seriesType='daily',
            forecastHorizon=self.forecastHorizon)

        for fd in range(self.forecastHorizon):
            temp_df = copy.deepcopy(self.datasets[fd].data)
            # add nan for the day before since 1st window is 17 to 1 o'clock
            for new_row in temp_df.index[0:24]-timedelta(days=1):
                temp_df.loc[new_row] = np.nan
            temp_df.sort_index(inplace=True)
            temp_df = evt.timeseries.moving_average_daily_max(
                temp_df, availability_ratio)

            new_sim.datasets[fd].updateFromDataset(temp_df)

        return new_sim

    def filteredSeries(self, availability_ratio=0.75):
        """ Return Simulations object working on the filtered series.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available for a given hour to compute
            the daily filtered series for this hour in all days.

        Returns
        -------
        evaluator.Simulations
            Simulations object with seriesType = 'hourly' and with data
            corresponding to the computed filtered series.

        """
        if self.seriesType is not 'hourly':
            print("Object's attribute 'seriesType' must be 'hourly' " +
                  "to use this method !!!")
            return None

        new_sim = self.__class__(
            startDate=self.startDate, endDate=self.endDate,
            stations=self.stations, species=self.species,
            model=self.model, seriesType='hourly',
            forecastHorizon=self.forecastHorizon)
        for fd in range(self.forecastHorizon):
            new_sim.datasets[fd].updateFromDataset(
                evt.timeseries.filtered_series(
                    self.datasets[fd].data, availability_ratio))
        return new_sim

    def normalized_series(self):
        """ Return Simulations object with normalized series.

        This method normalizes each series of simulations by substracting
        its median and dividing by its interquartile range.

        Returns
        -------
        evaluator.Simulations
            Simulations object with seriesType = 'hourly' and with data
            corresponding to the computed normalized series.

        """
        new_sim = self.__class__(
            startDate=self.startDate, endDate=self.endDate,
            stations=self.stations, species=self.species,
            model=self.model, seriesType=self.seriesType,
            forecastHorizon=self.forecastHorizon)
        for fd in range(self.forecastHorizon):
            new_sim.datasets[fd].updateFromDataset(
                evt.timeseries.normalized_series(self.datasets[fd].data))
        return new_sim


def load(inputFilePath):
    """ evaluator.Evaluator object loader

    Load an evaluator.Evaluator object saved in binary format
    with evaluator.Evaluator.dump method.

    Parameters
    ----------
    inputFilePath : str
        Path of the binary file to load.

    Returns
    -------
        evaluator.Evaluator object

    """
    with open(inputFilePath, 'rb') as f:
        obj = pickle.load(f)
    if type(obj) == evt.evaluator.Evaluator:
        return obj
    else:
        raise evt.EvaltoolsError("*** {} not an evaluator.Evaluator " +
                                 "object !!!".format(inputFilePath))


class Evaluator(object):
    """ Class gathering observations and simulations of the studied case.

    An object of class :class:`~evaltools.evaluator.Simulations` will be
    specific to one species, one series type, one period and one model. This
    class contain several methods to compare simulations and observations.

    Parameters
    ----------
    observations : evaluator.Observations object
        Observations used to construct the Evaluator object.
    simulations : evaluator.Simulations object
        Simulations used to construct the Evaluator object.
    color : str
        Default color that will be used in plotting functions.

    """
    def __init__(self, observations, simulations, color='k'):
        if all([ds.data.empty for ds in simulations.datasets]):
            raise evt.EvaltoolsError("Trying to construct Evaluator object " +
                                     "with empty simulations !!!")
        if observations.dataset.data.empty:
            raise evt.EvaltoolsError("Trying to construct Evaluator object " +
                                     "with empty observations !!!")
        assert(observations.species == simulations.species), \
            "observations.species != simulations.species !!!"
        assert(observations.startDate == simulations.startDate), \
            "observations.startDate != simulations.startDate !!!"
        assert(observations.endDate == simulations.endDate), \
            "observations.endDate != simulations.endDate !!!"
        assert(observations.forecastHorizon >= simulations.forecastHorizon), \
            "observations.forecastHorizon < simulations.forecastHorizon !!!"
        assert(observations.seriesType == simulations.seriesType), \
            "observations.seriesType != simulations.seriesType !!!"

        # only keep stations shared by obs and sim
        stationsIdx = np.intersect1d(
            observations.stations.index,
            simulations.stations)
        stations = observations.stations.loc[stationsIdx]

        self.observations = Observations(
            species=observations.species,
            startDate=observations.startDate,
            endDate=observations.endDate,
            stations=stations,
            seriesType=observations.seriesType,
            forecastHorizon=simulations.forecastHorizon)
        self.observations.dataset.updateFromDataset(observations.dataset)

        self.simulations = Simulations.fromDataset(
            model=simulations.model,
            ds_list=simulations.datasets,
            stationsIdx=stationsIdx)

        # color attribute used in plotting function
        self.color = color

        # dictionary where to save score results
        self.results = {}

    @property
    def species(self):
        return self.observations.species

    @property
    def startDate(self):
        return self.observations.startDate

    @property
    def endDate(self):
        return self.observations.endDate

    @property
    def forecastHorizon(self):
        return self.simulations.forecastHorizon

    @property
    def seriesType(self):
        return self.observations.seriesType

    @property
    def model(self):
        return self.simulations.model

    @property
    def stations(self):
        return self.observations.stations

    @property
    def obsDF(self):
        return self.observations.dataset.data

    @property
    def simDF(self):
        return [self.simulations.datasets[fd].data
                for fd in range(self.forecastHorizon)]

    @classmethod
    def fromHDF(cls, startDate, endDate, species, model, stations,
                obs_hdf_path, sim_hdf_path, listing_path, forecastHorizon=1,
                correc_unit=1, color='k'):
        """ Class method used to construct an object from HDF5 files.

        The architecture of HDF5 files must be the one used by
        :class:`~evaltools.dataframe.Store`.

        Parameters
        ----------
        startDate : datetime.date
            Starting day of the studied period.
        endDate : datetime.date
            Ending day (included) of the studied period.
        species : str
            Species name (must be a key of the hdf file).
        model : str
            Name of the model.
        stations : 1D array of str
            List of stations to keep in the resulting object.
        obs_hdf_path : str
            Path of the input file. Input data can be splitted over
            two hdf files corresponding to two different years. In
            that case, the file path must contain '{year}'.
        sim_hdf_path : str
            Path of the input file. The file path must contain
            '{forecastDay}' instead of the forecast day of the file.
            Input data can be splitted over two hdf files
            corresponding to two different years. In that case, the
            file path must contain '{year}'.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.
        color : str
            Default color that will be used in plotting functions.
        **kwargs :
            These parameters (like 'sep', 'sub_list', ...) will be
            passed to evaltools.utils.read_listing().

        Returns
        -------
            evaltools.evaluator.Evaluator object

        """
        observations = Observations.fromHDF(
            startDate, endDate, species, obs_hdf_path, listing_path,
            forecastHorizon=forecastHorizon, correc_unit=correc_unit,
            **kwargs)
        simulations = Simulations.fromHDF(
            startDate, endDate, species, model, stations, sim_hdf_path,
            listing_path, forecastHorizon=forecastHorizon,
            correc_unit=correc_unit, **kwargs)
        return cls(observations, simulations, color=color)

    def dump(self, outputFilePath):
        """ evaluator.Evaluator object loader

        Save an evaluator.Evaluator object in binary format.
        Use evaltools.evaluator.load function to get the object back.

        Parameters
        ----------
        outputFilePath : str
            Path of the output binary file.

        """
        with open(outputFilePath, 'wb') as f:
            pickle.dump(self, f)

    def summary(self):
        """ Print a summary of the object. """
        print("Model : {0}".format(self.model))
        print("Species : {0}".format(self.species))
        print("Period : {0} - {1}".format(
            self.startDate.strftime('%Y%m%d'),
            self.endDate.strftime('%Y%m%d')))
        print("Forecast horizon : {0}".format(self.forecastHorizon))

    def subPeriod(self, startDate, endDate):
        """ Build a new Evaluator object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Evaluator object.

        """
        return self.__class__(
            self.observations.subPeriod(startDate, endDate),
            self.simulations.subPeriod(startDate, endDate),
            color=self.color)

    def stationSubList(self, station_list, inplace=True):
        """ Drop stations not contained in the given list.

        Parameters
        ----------
        station_list : 1D array of str.
            List of stations to keep.
        inplace : bool
            If True, the Evaluator object is modified inplace,
            else, a new object is returned.

        """
        if inplace is True:
            obj = self
        else:
            obj = copy.deepcopy(self)
        obj.results = {}

        # check if station_list is included in self.stations.index
        stations = np.intersect1d(self.stations.index, station_list)
        missing_stations = np.array(station_list)[~np.in1d(station_list,
                                                           stations)]
        if missing_stations.shape[0] > 0:
            print(("Warning: stations {} from station_list were not found " +
                   "in the original object !!!").format(missing_stations))

        # drop unsought stations
        obj.observations.stations = obj.observations.stations.loc[stations]
        obj.observations.dataset.data = \
            obj.observations.dataset.data[stations]
        for fd in range(obj.forecastHorizon):
            obj.simulations.datasets[fd].data = \
                obj.simulations.datasets[fd].data[stations]

        if inplace is not True:
            return obj

    def subArea(self, minLon, maxLon, minLat, maxLat, inplace=True):
        """ Drop stations not contained within the given lat/lon boundaries.

        Parameters
        ----------
        minLon, maxLon, minLat, maxLat : scalars
            Lat/lon boundaries.
        inplace : bool
            If True, the Evaluator object is modified inplace,
            else, a new object is returned.

        """
        kept_stations_idx = reduce(
            np.logical_and,
            [self.stations['lat'] >= minLat, self.stations['lat'] <= maxLat,
             self.stations['lon'] >= minLon, self.stations['lon'] <= maxLon])
        if inplace is True:
            self.stationSubList(
                self.stations.index[kept_stations_idx],
                inplace=inplace)
        else:
            return self.stationSubList(
                self.stations.index[kept_stations_idx],
                inplace=inplace)

    def selectCountries(self, countries, inplace=True):
        """ Only keep stations within certain countries.

        This method assumes that the first caracters of a station code
        refer to the station country.

        Parameters
        ----------
        countries : 1D array of str.
            List of first letters of station codes to keep
            (e.g. ['FRA', 'IT', 'AUT', 'ES']).
        inplace : bool
            If True, the Evaluator object is modified inplace,
            else, a new object is returned.

        """
        kept_stations = [sta for sta in self.stations.index
                         if any([re.match(cn, sta) for cn in countries])]
        if inplace is True:
            self.stationSubList(kept_stations, inplace=inplace)
        else:
            return self.stationSubList(kept_stations, inplace=inplace)

    def time_series_to_csv(self, obsOutputFile, simOutputFile):
        """ Save timeseries dataframes as csv.

        Parameters
        ----------
        obsOutputFile : str
            File path where to save observed timeseries.
        simOutputFile : str
            File path where to save simulated timeseries. The path name
            must contain {forecastDay} instead of the forecast day number.

        """
        self.obsDF.to_csv(obsOutputFile, sep=',', na_rep='nan',
                          float_format='%g', header=True, index=True,
                          date_format='%Y%m%d%H')
        for forecast_day in range(self.forecastHorizon):
            self.simDF[forecast_day].to_csv(
                simOutputFile.format(forecastDay=forecast_day),
                sep=',', na_rep='nan', float_format='%g', header=True,
                index=True, date_format='%Y%m%d%H')

    def get_obs(self, forecastDay, start_end=None):
        """ Get observations according to the forecast day.

        Parameters
        ----------
        forecastDay : int
            Forecast day for which to get observations.
        start_end : None or list of two datetime.date objects
            Dates between which getting the data.

        Returns
        -------
        pandas.DataFrame
            Observations for the time period corresponding to
            the forecast day.

        """
        if start_end is not None:
            start = start_end[0] + timedelta(days=forecastDay)
            end = start_end[1] + timedelta(days=forecastDay)
        else:
            start = self.startDate + timedelta(days=forecastDay)
            end = self.endDate + timedelta(days=forecastDay)

        idx = ((self.obsDF.index.date >= start) &
               (self.obsDF.index.date <= end))

        return self.obsDF.loc[idx]

    def get_sim(self, forecastDay, start_end=None):
        """ Get simulations according to the forecast day.

        Parameters
        ----------
        forecastDay : int
            Forecast day for which to get simulations.
        start_end : None or list of two datetime.date objects
            Dates between which getting the data.

        Returns
        -------
        pandas.DataFrame
            Simulations corresponding to the forecast day.

        """
        if start_end is not None:
            start = start_end[0] + timedelta(days=forecastDay)
            end = start_end[1] + timedelta(days=forecastDay)

            idx = ((self.simDF[forecastDay].index.date >= start) &
                   (self.simDF[forecastDay].index.date <= end))

            return self.simDF[forecastDay].loc[idx]

        return self.simDF[forecastDay]

    def stationScores(self, score_list, outputFile=None,
                      availability_ratio=0.75):
        """ Compute scores per station for each forecast day.

        This method may not be appropriate for hourly time steps data.
        You can also checkout FTscores method.

        Parameters
        ----------
        score_list : list of str
            List of computed scores.
        outputFile : str
            File where to save the result. The file name must contain
            {forecastDay} instead of the forecast day number. If None,
            result is not saved in csv.
        availability_ratio : float
            Minimal rate of data available on the period required per forecast
            day to compute the scores.

        Returns
        -------
        dictionary
            Dictionary with one key per forecast day ('D0', 'D1', ...)
            containing pandas.DataFrames with station names as index and one
            column per computed score.

        """
        res = {}
        for forecast_day in range(self.forecastHorizon):
            obs = self.get_obs(forecast_day)
            sim = self.get_sim(forecast_day)
            statsDFsta = evt.scores.stats2D(
                obs, sim, score_list, axis=0,
                threshold=float(availability_ratio), keep_nan=True)
            if outputFile is not None:
                statsDFsta.to_csv(outputFile.format(forecastDay=forecast_day),
                                  sep=' ', na_rep='nan', float_format='%g',
                                  header=True, index=True,
                                  date_format='%Y%m%d%H')
            res["D{}".format(str(forecast_day))] = statsDFsta
        return res

    def quarterlyMedianScore(self, filePath, score='RMSE', forecastDay=0,
                             availability_ratio=0.75):
        """ Compute median on station scores.

        Parameters
        ----------
        filePath : str
            File where to find previous values and save the result.
        score : str
            The score to process.
        forecastDay : int
            Integer corresponding to the chosen forecastDay used for
            computation.
        availability_ratio : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores.

        Returns
        -------
            The median for the computed score.

        """
        stats = self.stationScores(score_list=[score],
                                   availability_ratio=availability_ratio
                                   )['D{}'.format(forecastDay)][score]

        if isfile(filePath):
            df = pd.read_csv(filePath, sep=' ', index_col=0)
        else:
            df = pd.DataFrame(columns=[score])

        quarter = evt.quarter.Quarter(self.startDate, self.endDate)
        idx = quarter.__repr__()

        res = np.nanmedian(stats)
        df.loc[idx] = res
        df.to_csv(filePath, sep=' ', na_rep='nan', float_format='%g',
                  header=True, index=True, date_format='%Y%m%d%H')
        return res

    def timeScores(self, score_list, outputFile=None, min_nb_sta=10):
        """ Compute scores per time.

        Compatible both for hourly and daily series types.

        Parameters
        ----------
        score_list : list of str
            List of computed scores.
        outputFile : str.
            File where to save the result. The file name must contain
            {forecastDay} instead of the forecast day number. If None,
            result is not saved in csv.
        min_nb_sta : int
            Minimal number of station values available in both obs and
            sim required per datetime to compute the scores.

        Returns
        -------
        dictionary
            Dictionary with one key per forecast day ('D0', 'D1', ...)
            containing pandas.DataFrames with datetime index and one column
            per computed score.

        """
        try:
            # test if these score are already computed
            assert (self.results['timeScores']['args'] ==
                    {'score_list': score_list, 'min_nb_sta': min_nb_sta})
            res = self.results['timeScores']['result']

        except (AssertionError, KeyError):
            # print("computing timescores")
            res = {}
            for fd in range(self.forecastHorizon):
                obs = self.get_obs(fd)
                sim = self.get_sim(fd)
                statsDFsta = evt.scores.stats2D(obs, sim, score_list, axis=1,
                                                threshold=int(min_nb_sta))
                if outputFile is not None:
                    statsDFsta.to_csv(outputFile.format(forecastDay=fd),
                                      sep=' ', na_rep='nan', header=True,
                                      float_format='%g', index=True,
                                      date_format='%Y%m%d%H')
                res["D{}".format(str(fd))] = statsDFsta
            self.results['timeScores'] = {'args': {'score_list': score_list,
                                                   'min_nb_sta': min_nb_sta},
                                          'result': res}

        return res

    def meanTimeScores(self, score_list, outputFile=None, min_nb_sta=10,
                       availability_ratio=0.75):
        """ Compute the mean of time scores for each forecast time.

        Parameters
        ----------
        score_list : list of str
            List of computed scores.
        outputFile : str
            File where to save the result. The file name must contain
            {forecastDay} instead of the forecast day number. If None,
            result is not saved in csv.
        min_nb_sta : int
            Minimal number of station values available in both obs and sim
            required per datetime to compute the scores (before applying mean
            for each forecast time).
        availability_ratio : float
            Minimal rate of data (computed scores per time) available on the
            period required per forecast time to compute the mean scores.

        Returns
        -------
        dictionary
            Dictionary, with one key per score, that contains lists of the
            means for each forecast time. For example if the forecast horizon
            is 4, the lists will be of length 96.

        """
        if self.seriesType not in ['hourly']:
            raise evt.EvaltoolsError("Object's attribute 'seriesType' must " +
                                     "be 'hourly' to use this method !!!")

        # scores computing
        stats = self.timeScores(score_list, min_nb_sta=min_nb_sta)

        res = {}
        nb_days = (self.endDate - self.startDate).days + 1
        for sc in score_list:
            Y = []
            for fd in range(self.forecastHorizon):
                matrix = stats["D{}".format(str(fd))][sc].values.reshape(
                    nb_days, 24)
                idx = (np.sum(np.isnan(matrix), axis=0) >=
                       (availability_ratio*matrix.shape[0]))
                with np.warnings.catch_warnings():
                    np.warnings.filterwarnings('ignore',
                                               'Mean of empty slice')
                    means = np.nanmean(matrix, axis=0)
                means[idx] = np.nan
                Y.extend(list(means))
            res[sc] = Y
        if outputFile is not None:
            pd.DataFrame(res).to_csv(outputFile, sep=' ', na_rep='nan',
                                     float_format='%g', header=True,
                                     index=True, date_format='%Y%m%d%H')

        return res

    def FTscores(self, score_list, availability_ratio=0.75, outputFile=None):
        """ Compute forecast time scores for each station.

        Only available for hourly time step data.

        Parameters
        ----------
        score_list : list of str
            List of scores to compute.
        availability_ratio : float
            Minimal rate of data available on the period required per
            forecast time to compute the scores for each station.
        outputFile : str
            File where to save the result. If None, result is not saved in
            csv. The file name must contain {score} instead of the score
            name. A file in created for each score.

        Returns
        -------
        dictionary
            Dictionary, with one key per score, that contains
            pandas.DataFrame with one row per station and one column per
            forecast time except the first two columns that are latitude
            and longitude.

        """
        try:
            assert (self.results['FTscores']['args'] ==
                    {'score_list': score_list,
                     'availability_ratio': availability_ratio})
            res = self.results['FTscores']['result']

        except (AssertionError, KeyError):
            if self.seriesType not in ['hourly']:
                print("Object's attribute 'seriesType' must be 'hourly' " +
                      "to use this method !!!")
                return None
            res = {}
            for sc in score_list:
                res[sc] = self.stations[['lat', 'lon']]
                res[sc].columns = pd.MultiIndex.from_product(
                    [['coord'], ['lat', 'lon']])
            for day in range(self.forecastHorizon):
                obs = self.get_obs(day)
                sim = self.get_sim(day)
                dailyStats = evt.scores.FTstats(obs, sim, score_list,
                                                availability_ratio, day)
                midx = pd.MultiIndex.from_product(
                    [['D{}'.format(day)], range(24)])
                for sc in score_list:
                    dailyStats[sc].columns = midx
                    res[sc] = pd.concat([res[sc], dailyStats[sc]],
                                        axis=1, sort=False)
            if outputFile is not None:
                for sc in score_list:
                    file_path = outputFile.format(score=sc)
                    res[sc].insert(0, ('station', 'station'), res[sc].index)
                    res[sc].to_csv(file_path, sep=',', na_rep='nan',
                                   float_format='%g', header=True,
                                   index=False, date_format='%Y%m%d%H')
            self.results['FTscores'] = {
                'args': {'score_list': score_list,
                         'availability_ratio': availability_ratio},
                'result': res}

        return res

    def medianStationScores(self, score_list, availability_ratio=0.75,
                            min_nb_sta=10, outputFile=None):
        """ Compute median station scores for each forecast time.

        Only available for hourly time step data.

        Parameters
        ----------
        score_list : list of str
            List of scores to compute.
        availability_ratio : float
            Minimal rate of data available on the period required per
            forecast time to compute the scores for each station.
        min_nb_sta : int
            Minimal number of station values available in both obs and sim
            required to compute the median.
        outputFile : str
            File where to save the result. If None, result is not saved in
            csv. The file name must contain {score} instead of the score
            name. A file is created for each score.

        Returns
        -------
        pandas.DataFrame
            DataFrame with one row per forecast time and one column
            per score.

        """
        df_dict = self.FTscores(score_list,
                                availability_ratio=availability_ratio,
                                outputFile=None)

        med_scores = {}
        for sc in score_list:
            # [2:] to remove lat lon
            med_scores[sc] = np.nanmedian(df_dict[sc], axis=0)[2:]

        res = pd.DataFrame(med_scores, index=range(self.forecastHorizon*24))

        idx_not_enough_values = np.sum(~list(df_dict.values())[0].isna(),
                                       axis=0)[2:].values < min_nb_sta
        res[idx_not_enough_values] = np.nan

        if outputFile is not None:
            pd.DataFrame(res).to_csv(outputFile, sep=' ', na_rep='nan',
                                     float_format='%g', header=True,
                                     index=True, date_format='%Y%m%d%H')

        return res

    def FDscores(self, score_list, forecast_days='all', threshold=0.75,
                 outputFile=None):
        """ Compute forecast day scores for the model.

        Scores are computed using all data available for a given forecast day
        (no mean or median is taken).

        Parameters
        ----------
        score_list : list of str
            List of scores to compute.
        forecast_days : 'all' or list of int
            Forcast days used for computation. The returned DataFrame will
            contain one row per forcast day.
        threshold : int or float
            Minimal number (if type(threshold) is int) or minimal rate
            (if type(threshold) is float) of data available in both obs and
            sim required to compute the scores.
        outputFile : str
            File where to save the result. If None, result is not saved
            in csv.

        Returns
        -------
        pandas.DataFrame
            DataFrame with one column per score and one row per
            forecast day.

        """
        res = pd.DataFrame()
        if forecast_days is 'all':
            forecast_days = range(self.forecastHorizon)

        for day in forecast_days:
            obs = self.get_obs(day).values.flatten()
            sim = self.get_sim(day).values.flatten()
            # define threshold in terms of number of obs instead of ratio
            if type(threshold) is float:
                if threshold >= 1.:
                    print("*** Warning: threshold of type float >= 1")
                thresh = threshold*obs.shape[0]
            df = pd.DataFrame(
                evt.scores.stats2D_numpy(obs.reshape(obs.size, 1),
                                         sim.reshape(sim.size, 1),
                                         score_list,
                                         threshold=thresh),
                index=["D{}".format(day)],
                columns=score_list)
            res = res.append(df)

        if outputFile is not None:
            res.to_csv(outputFile, sep=' ', na_rep='nan',
                       float_format='%g', header=True, index=True)
        return res

    def dailyMean(self, availability_ratio=0.75):
        """ Build Evaluator object working on daily mean.

        This method computes the daily mean of observations and simulations
        of the current object.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily mean.

        Returns
        -------
        evaluator.Evaluator
            Evaluator object with seriesType = 'daily' and with data
            corresponding to the computed daily mean.

        """
        return self.__class__(self.observations.dailyMean(availability_ratio),
                              self.simulations.dailyMean(availability_ratio),
                              color=self.color)

    def dailyMax(self, availability_ratio=0.75):
        """ Return Evaluator object working on daily maximum.

        This method compute the daily maximum of observations and simulations
        of the current object and returns a new Evaluator object with
        seriesType = 'daily' and with data corresponding to the computed
        daily maxima.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily maximum.

        Returns
        -------
        evaluator.Evaluator
            Evaluator object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum.

        """
        return self.__class__(self.observations.dailyMax(availability_ratio),
                              self.simulations.dailyMax(availability_ratio),
                              color=self.color)

    def movingAverageDailyMax(self, availability_ratio=0.75):
        """ Compute the daily maximum of the moving average.

        This method compute the daily maximum of the moving average
        for observations and simulations of the current object and returns
        a new Evaluator object.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available to compute the average for a
            given 8-hour window, and also minimal rate of values available
            for a given day to compute the daily maximum of the moving
            average. For example with availability_ratio=0.75, a daily
            maximum eight hours average can only be calculated if 18 eight
            hours average are available each of which requires 6 hourly
            values to be available.

        Returns
        -------
        evaluator.Evaluator
            Evaluator object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum of the moving average.

        """
        return self.__class__(
            self.observations.movingAverageDailyMax(availability_ratio),
            self.simulations.movingAverageDailyMax(availability_ratio),
            color=self.color)

    def filteredSeries(self, availability_ratio=0.75):
        """ Return Evaluator object working on the filtered series.

        This method compute the filtered series of observations and
        simulations of the current object and returns a new Evaluator object
        with seriesType = 'hourly' and with data corresponding to the computed
        filtered series.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available for a given hour to compute the
            daily filtered series for this hour in all days.

        Returns
        -------
        evaluator.Evaluator
            Evaluator object with seriesType = 'hourly' and with data
            corresponding to the computed filtered series.

        """
        return self.__class__(
            self.observations.filteredSeries(availability_ratio),
            self.simulations.filteredSeries(availability_ratio),
            color=self.color)

    def normalized_series(self):
        """ Return Evaluator object with normalized series.

        This method normalizes each series of observations and simulations
        by substracting its median and dividing by its interquartile range.

        Returns
        -------
        evaluator.Evaluator
            Evaluator object with seriesType = 'hourly' and with data
            corresponding to the computed normalized series.

        """
        return self.__class__(self.observations.normalized_series(),
                              self.simulations.normalized_series(),
                              color=self.color)

    def contingency_table(self, threshold, outputFile):
        """ Contingency table.

        The table is computed from daily data for each forecast day.
        Before computed the table, values for every stations and every
        days of the period are concatenated. Tables corresponding to
        the different forecast days are stored in the same file.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        outputFile : str
            File where to save the result.

        """
        if self.seriesType not in ['daily']:
            print("*** Object's attribute 'seriesType' must be 'daily' " +
                  "to use this method !!!")
            return

        with open(outputFile, 'w') as f:
            for fd in range(self.forecastHorizon):
                obs = self.get_obs(fd)
                obs = obs.values.flatten()
                sim = self.get_sim(fd)
                sim = sim.values.flatten()
                res = evt.scores.contingency_table(obs, sim, thr=threshold)
                res = pd.DataFrame(
                    res,
                    index=["obs>s", "obs<s", "total"],
                    columns=["sim>s", "sim<s", "total"]).astype(int)
                f.write("D{}\n".format(fd) + res.T.to_string() + '\n\n')

    def obs_exceedances(self, threshold, outputFile):
        """ Look for exceedances in observed time series.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        outputFile : str
            File where to save the result.

        """
        ts = self.obsDF.stack()
        idx = ts > threshold
        res = ts.loc[idx]
        with open(outputFile, 'w') as f:
            f.write(res.to_string())

    def sim_exceedances(self, threshold, outputFile):
        """ Look for exceedances in simulated time series.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        outputFile : str
            File where to save the result. The file name must contain
            {forecastDay} instead of the forecast day number.

        """
        for fd in range(self.forecastHorizon):
            ts = self.get_sim(fd).stack()
            idx = ts > threshold
            res = ts.loc[idx]
            with open(outputFile.format(forecastDay=fd), 'w') as f:
                f.write(res.to_string())

    def fairmodeBenchmark(self, targetFile=None, summaryFile=None,
                          outputCSV=None, availability_ratio=0.75, label=None,
                          targetTitle=None, summaryTitle=None, color=None,
                          file_formats=['png'], forecastDay=0):
        """ Plot FAIRMODE target and summary diagrams.

        Concentration values must be in µg/m^3. Supported species are 'o3',
        'no2', 'pm10' and 'pm2p5'.

        Parameters
        ----------
        targetFile :  str or None
            File where to save the target diagram (without extension).
            If None, the figure is shown in a popping window.
        summaryFile :  str or None
            File where to save the summary diagram (without extension).
            If None, the figure is shown in a popping window.
        outputCSV : str or None
            File where to save the target data.
        availability_ratio : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores for each station.
        label : str
            Label for the legend.
        targetTitle : str
            Target diagram title.
        summaryTitle : str
            Summary diagram title.
        color : str
        file_formats : list of str
            List of file extensions.
        forecastDay : int
            Forecast day used to compute the two diagrams.

        """
        if self.seriesType is not 'hourly':
            print("Object's attribute 'seriesType' must be 'hourly' to " +
                  "use this method !!!")
            return None

        # parameters specific to the species
        if self.species == 'pm10':
            obj = evt.fairmode.FairmodeEvaluator(
                self.observations.dailyMean(availability_ratio),
                self.simulations.dailyMean(availability_ratio),
                color=self.color,
                threshold=50.,
                U=0.28,
                alpha=0.13,
                RV=50,
                perc=0.901,
                Np=30.,
                Nnp=0.25,
                beta=2.
                )
        elif self.species == 'pm2p5':
            obj = evt.fairmode.FairmodeEvaluator(
                self.observations.dailyMean(availability_ratio),
                self.simulations.dailyMean(availability_ratio),
                color=self.color,
                threshold=25,
                U=0.36,
                alpha=0.3,
                RV=25,
                perc=0.901,
                Np=30.,
                Nnp=0.25,
                beta=2.)
        elif self.species == 'no2':
            obj = evt.fairmode.FairmodeEvaluator(
                self.observations,
                self.simulations,
                color=self.color,
                threshold=200.,
                U=0.24,
                alpha=0.2,
                RV=200,
                perc=0.998,
                Np=5.2,
                Nnp=5.5,
                beta=2.)
        elif self.species == 'o3':
            obj = evt.fairmode.FairmodeEvaluator(
                self.observations.movingAverageDailyMax(availability_ratio),
                self.simulations.movingAverageDailyMax(availability_ratio),
                color=self.color,
                threshold=120.,
                U=0.18,
                alpha=0.79,
                RV=120,
                perc=0.929,
                Np=11.,
                Nnp=3.,
                beta=2.)
        else:
            print("Species {} not supported by Fairmode's metrics.".format(
                self.species))
            return None

        species_names = {'o3': r'Surface ozone 8h moving average daily ' +
                               r'maximum [$\mu g/m^3$]',
                         'no2': r'Surface nitrogen dioxide hourly mean ' +
                                r'[$\mu g/m^3$]',
                         'pm10': r'Surface PM10 aerosol daily mean ' +
                                 r'[$\mu g/m^3$]',
                         'pm2p5': r'Surface PM2.5 aerosol daily mean ' +
                                  r'[$\mu g/m^3$]'}

        # target
        if targetTitle is None:
            targetTitle = ("\n{spe} \n{startDate} 00UTC to {endDate} " +
                           "00UTC").format(spe=species_names[self.species],
                                           startDate=self.startDate,
                                           endDate=self.endDate)
        obj.plot_targetDiagram(
            availability_ratio=availability_ratio, forecastDay=forecastDay,
            label=label, color=color, title=targetTitle,
            outputFile=targetFile, file_formats=file_formats,
            outputCSV=outputCSV, list_stations_ouside_target=True)

        # summary
        if summaryTitle is None:
            summaryTitle = ("{spe} \n{startDate} 00UTC to {endDate} " +
                            "00UTC").format(spe=species_names[self.species],
                                            startDate=self.startDate,
                                            endDate=self.endDate)
        obj.plot_summary(
            availability_ratio=availability_ratio, forecastDay=forecastDay,
            outputFile=summaryFile, file_formats=file_formats,
            title=summaryTitle)
